use actix_http::header::CONTENT_DISPOSITION;
use actix_web::{http::header::ContentDisposition, web, HttpResponse, Responder};
use chrono::{Datelike, NaiveDate};
use rust_xlsxwriter::{workbook::Workbook, Format};
use serde::Deserialize;
use sqlx::{query, PgPool};

use crate::{
    models::{Role, User},
    utils::{ApplicationError, DateTimeFormat},
};

#[derive(Deserialize)]
struct ExportQuery {
    start: NaiveDate,
    end: NaiveDate,
    area: Option<i32>,
}

#[derive(PartialEq)]
struct EventExportEntry {
    date: String,
    weekday: String,
    start_time: String,
    end_time: String,
    hours: f32,
    location: String,
    name: String,
    assigned_name: Option<String>,
    assigned_function: Option<String>,
}

async fn read(pool: &PgPool) -> Result<Vec<EventExportEntry>, ApplicationError> {
    let results = query!(
        "select
	event.starttimestamp,
	event.endtimestamp,
	event.amountofposten,
        event.voluntaryfuehrungsassistent,
        event.voluntarywachhabender,
	location.name as locationName,
	event.name as eventName,
	array (
	select
		user_.name || ' --- ' || assignment.function
	from
		assignment
	join availability on
		assignment.availabilityid = availability.id
	join user_ on
		availability.userid = user_.id
	where
		assignment.eventId = event.id) as assignments,
		array (
	select
		vehicle.station || ' ' || vehicle.radiocallname
	from
		vehicleassignment
	join vehicle on
		vehicleassignment.vehicleId = vehicle.id
	where
		vehicleassignment.eventId = event.id) as vehicles
from
	event
join location on
	event.locationId = location.id
        order by event.starttimestamp"
    )
    .fetch_all(pool)
    .await?;

    let mut entries = Vec::new();

    for r in results {
        let date = r
            .starttimestamp
            .date_naive()
            .format(DateTimeFormat::DayMonthYear.into());
        let weekday = r.starttimestamp.date_naive().weekday();
        let start_time = r
            .starttimestamp
            .time()
            .format(DateTimeFormat::HourMinute.into());
        let end_time = r
            .endtimestamp
            .time()
            .format(DateTimeFormat::HourMinute.into());
        let hours = (r.endtimestamp - r.starttimestamp).as_seconds_f32() / 3600.0 + 1.0;
        let location = r.locationname;
        let name = r.eventname;

        let mut event_new: Vec<EventExportEntry> = Vec::new();

        let mut assigned_posten: i16 = 0;
        let mut assigned_wachhabender = false;
        let mut assigned_fuehrungassistent = false;

        if let Some(assignments) = r.assignments {
            for a in assignments {
                let (assigned_name, assigned_function) = a.split_once("---").unwrap();
                let assigned_function = match assigned_function.trim() {
                    "posten" => {
                        assigned_posten += 1;
                        "PO"
                    }
                    "wachhabender" => {
                        assigned_wachhabender = true;
                        "WH"
                    }
                    "fuehrungsassistent" => {
                        assigned_fuehrungassistent = true;
                        "FüAss"
                    }
                    _ => assigned_function.trim(),
                };

                event_new.push(EventExportEntry {
                    date: date.to_string(),
                    weekday: weekday.to_string(),
                    start_time: start_time.to_string(),
                    end_time: end_time.to_string(),
                    hours,
                    location: location.to_string(),
                    name: name.to_string(),
                    assigned_name: Some(assigned_name.trim().to_string()),
                    assigned_function: Some(assigned_function.to_string()),
                });
            }
        }

        if !assigned_wachhabender {
            let function = if r.voluntarywachhabender {
                "WH"
            } else {
                "BF-WH"
            };

            event_new.push(EventExportEntry {
                date: date.to_string(),
                weekday: weekday.to_string(),
                start_time: start_time.to_string(),
                end_time: end_time.to_string(),
                hours,
                location: location.to_string(),
                name: name.to_string(),
                assigned_name: None,
                assigned_function: Some(function.to_string()),
            });
        }

        if !assigned_fuehrungassistent && r.voluntaryfuehrungsassistent {
            event_new.push(EventExportEntry {
                date: date.to_string(),
                weekday: weekday.to_string(),
                start_time: start_time.to_string(),
                end_time: end_time.to_string(),
                hours,
                location: location.to_string(),
                name: name.to_string(),
                assigned_name: None,
                assigned_function: Some("FüAss".to_string()),
            });
        }

        if assigned_posten < r.amountofposten {
            for _ in 0..(r.amountofposten - assigned_posten) {
                event_new.push(EventExportEntry {
                    date: date.to_string(),
                    weekday: weekday.to_string(),
                    start_time: start_time.to_string(),
                    end_time: end_time.to_string(),
                    hours,
                    location: location.to_string(),
                    name: name.to_string(),
                    assigned_name: None,
                    assigned_function: Some("PO".to_string()),
                });
            }
        }

        if let Some(vehicles) = r.vehicles {
            for v in vehicles {
                event_new.push(EventExportEntry {
                    date: date.to_string(),
                    weekday: weekday.to_string(),
                    start_time: start_time.to_string(),
                    end_time: end_time.to_string(),
                    hours,
                    location: location.to_string(),
                    name: name.to_string(),
                    assigned_name: Some(v.to_string()),
                    assigned_function: Some(String::from("Fzg")),
                });
            }
        }

        entries.append(&mut event_new);
    }

    Ok(entries)
}

#[actix_web::get("/export/eventsdata")]
pub async fn get(
    pool: web::Data<PgPool>,
    user: web::ReqData<User>,
    query: web::Query<ExportQuery>,
) -> Result<impl Responder, ApplicationError> {
    if user.role != Role::Admin && user.role != Role::AreaManager {
        return Err(ApplicationError::Unauthorized);
    }

    let entries = read(pool.get_ref()).await?;

    let mut workbook = Workbook::new();
    let worksheet = workbook.add_worksheet();
    // let time_format = Format::new();
    // time_format.set_num_format(num_format)

    const HEADER: [&str; 10] = [
        "Datum",
        "Wochentag",
        "Beginn",
        "Ende",
        "Stunden",
        "Ort",
        "VA",
        "Namen",
        "Fkt",
        "Reserve",
    ];
    worksheet.write_row(2, 0, HEADER).unwrap();

    for (i, entry) in entries.iter().enumerate() {
        let i = (i + 3) as u32;

        worksheet.write(i, 0, &entry.date).unwrap();
        worksheet.write(i, 1, &entry.weekday).unwrap();
        worksheet.write(i, 2, &entry.start_time).unwrap();
        worksheet.write(i, 3, &entry.end_time).unwrap();
        worksheet.write(i, 4, entry.hours).unwrap();
        worksheet
            .write(i, 4, format!("{:.1}", entry.hours))
            .unwrap();
        worksheet.write(i, 5, &entry.location).unwrap();
        worksheet.write(i, 6, &entry.name).unwrap();
        worksheet.write(i, 7, entry.assigned_name.as_ref()).unwrap();
        worksheet
            .write(i, 8, entry.assigned_function.as_ref())
            .unwrap();
    }

    worksheet.autofit();

    let buffer = workbook.save_to_buffer().unwrap();

    return Ok(HttpResponse::Ok()
        .content_type("application/vnd.openxmlformats-officedocument.spreadsheetml.sheet")
        .insert_header((
            CONTENT_DISPOSITION,
            ContentDisposition::attachment("export.xlsx"),
        ))
        .body(buffer));
}
